
/*****************************************************************************/
/*                                                                           */
/*  THE KHE HIGH SCHOOL TIMETABLING ENGINE                                   */
/*  COPYRIGHT (C) 2010 Jeffrey H. Kingston                                   */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         khe_sr_dynamic_vlsn.c                                      */
/*  DESCRIPTION:  Dynamic resource VLSN solving                              */
/*                                                                           */
/*****************************************************************************/
#include "khe_solvers.h"
#include <limits.h>
#define is_digit(x) ((x) >= '0' && (x) <= '9')
#define bool_show(x) ((x) ? "true" : "false")

#define DEBUG1 1
#define DEBUG2 1
#define DEBUG3 1
#define DEBUG4 0
#define DEBUG5 0
#define DEBUG6 0


/*****************************************************************************/
/*                                                                           */
/* Submodule "type declarations"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_INTERVAL - one interval                                              */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_interval_rec {
  int first;
  int last;
} KHE_INTERVAL;

typedef HA_ARRAY(KHE_INTERVAL) ARRAY_KHE_INTERVAL;


/*****************************************************************************/
/*                                                                           */
/*  KHE_INTERVAL_SET - a set of intervals                                    */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_interval_set_rec {
  ARRAY_KHE_INTERVAL	intervals;		/* the intervals */
} *KHE_INTERVAL_SET;

typedef HA_ARRAY(KHE_INTERVAL_SET) ARRAY_KHE_INTERVAL_SET;


/*****************************************************************************/
/*                                                                           */
/*  KHE_DAY_SELECT - a sequence of interval sets                             */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_day_select_rec {
  int				random_len;		/* if random:x       */
  ARRAY_KHE_INTERVAL_SET	interval_sets;
} *KHE_DAY_SELECT;

typedef HA_ARRAY(KHE_RESOURCE) ARRAY_KHE_RESOURCE;


/*****************************************************************************/
/*                                                                           */
/*  Statistics                                                               */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_dim_dom_kind_item_rec *KHE_DIM_DOM_KIND_ITEM;

typedef struct khe_drs_stats_day_index_rec {
  int				day_index;
  int				made_count;
  int				table_size;
  float				running_time;
} *KHE_DRS_STATS_DAY_INDEX;

typedef HA_ARRAY(KHE_DRS_STATS_DAY_INDEX) ARRAY_KHE_DRS_STATS_DAY_INDEX;

typedef struct khe_drs_stats_resource_count_rec {
  int				resource_count;
  ARRAY_KHE_DRS_STATS_DAY_INDEX	days;
  KHE_COST			cost;
} *KHE_DRS_STATS_RESOURCE_COUNT;

typedef HA_ARRAY(KHE_DRS_STATS_RESOURCE_COUNT)
  ARRAY_KHE_DRS_STATS_RESOURCE_COUNT;

typedef struct khe_drs_stats_dom_kind_rec {
  KHE_DIM_DOM_KIND_ITEM			dom_kind_item;
  ARRAY_KHE_DRS_STATS_RESOURCE_COUNT	resource_counts;
} *KHE_DRS_STATS_DOM_KIND;

typedef HA_ARRAY(KHE_DRS_STATS_DOM_KIND) ARRAY_KHE_DRS_STATS_DOM_KIND;

typedef struct khe_drs_stats_rec {
  ARRAY_KHE_DRS_STATS_DOM_KIND	dom_kinds;
} *KHE_DRS_STATS;


/*****************************************************************************/
/*                                                                           */
/*  Testing                                                                  */
/*                                                                           */
/*****************************************************************************/

typedef enum {
  KHE_DIM_DOM_KIND_TAG,
  KHE_DIM_RESOURCE_COUNT_TAG,
  KHE_DIM_DAY_INDEX_TAG
} KHE_DIM_TAG;

typedef HA_ARRAY(KHE_DRS_DOM_KIND) ARRAY_KHE_DRS_DOM_KIND;

#define INHERIT_KHE_DIM						\
  KHE_DIM_TAG			tag;				\

typedef struct khe_dim_rec {
  INHERIT_KHE_DIM
} *KHE_DIM;

struct khe_dim_dom_kind_item_rec {
  bool				priqueue;
  bool				extra_selection;
  bool				expand_by_shifts;
  KHE_DRS_DOM_KIND		main_dom_kind;
  bool				cache;
  KHE_DRS_DOM_KIND		cache_dom_kind;
};

typedef HA_ARRAY(KHE_DIM_DOM_KIND_ITEM) ARRAY_KHE_DIM_DOM_KIND_ITEM;

typedef struct khe_dim_dom_kind_rec {
  INHERIT_KHE_DIM
  ARRAY_KHE_DIM_DOM_KIND_ITEM	items;
} *KHE_DIM_DOM_KIND;

typedef struct khe_dim_resource_count_rec {
  INHERIT_KHE_DIM
  HA_ARRAY_INT			resource_count_ranges;
  HA_ARRAY_INT			resource_count_singles;
} *KHE_DIM_RESOURCE_COUNT;

typedef HA_ARRAY(KHE_TIME_GROUP) ARRAY_KHE_TIME_GROUP;

typedef struct khe_dim_day_index_rec {
  INHERIT_KHE_DIM
  bool				output_whole_axis;
  ARRAY_KHE_TIME_GROUP		day_time_group_ranges;
  HA_ARRAY_INT			day_index_ranges;
  HA_ARRAY_INT			day_index_singles;
} *KHE_DIM_DAY_INDEX;

typedef HA_ARRAY(KHE_DIM) ARRAY_KHE_DIM;

typedef struct khe_drs_run_rec {
  HA_ARENA			arena;
  KHE_SOLN			soln;
  KHE_RESOURCE_TYPE		resource_type;
  ARRAY_KHE_RESOURCE		resources;
  ARRAY_KHE_DIM			dims;
  KHE_DIM_DOM_KIND		dom_kind_dim;
  KHE_DIM_RESOURCE_COUNT	resource_count_dim;
  KHE_DIM_DAY_INDEX		day_index_dim;
  KHE_DRS_STATS			stats;
} *KHE_DRS_RUN;


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_INTERVAL"                                                 */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_INTERVAL KheIntervalMake(int first, int last)                        */
/*                                                                           */
/*  Make a new interval object with these attributes.                        */
/*                                                                           */
/*****************************************************************************/

static KHE_INTERVAL KheIntervalMake(int first, int last)
{
  KHE_INTERVAL res;
  res.first = first;
  res.last = last;
  return res;
}

/*****************************************************************************/
/*                                                                           */
/*  char *KheReadInt(char *str, int *res)                                    */
/*                                                                           */
/*  Read an integer from str, returning a pointer to the first place past    */
/*  what was read.                                                           */
/*                                                                           */
/*  Implementation note.  We don't use "%n" to find the first place past     */
/*  what was read, because "man sscanf" says that the return value of        */
/*  sscanf is not clearly specified when "%n" is used.                       */
/*                                                                           */
/*****************************************************************************/

static char *KheReadInt(char *str, int *res)
{
  if( sscanf(str, "%d", res) != 1 )
    HnAbort("KheDynamicResourceVLSNSolve: error in rs_drs_day_select option");
  while( is_digit(*str) )
    str++;
  return str;
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheIntervalRead(char *str, KHE_INTERVAL *res)                      */
/*                                                                           */
/*  Read an interval from str, returning a pointer to the first place past   */
/*  what was read.                                                           */
/*                                                                           */
/*****************************************************************************/

static char *KheIntervalRead(char *str, KHE_INTERVAL *res)
{
  int first, last;
  str = KheReadInt(str, &first);
  if( *str != '-' )
    HnAbort("KheDynamicResourceVLSNSolve: error in rs_drs_day_select option");
  str = KheReadInt(str+1, &last);
  *res = KheIntervalMake(first, last);
  return str;
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_INTERVAL_SET"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_INTERVAL_SET KheIntervalSetMake(HA_ARENA a)                          */
/*                                                                           */
/*  Make a new, empty interval set object.                                   */
/*                                                                           */
/*****************************************************************************/

static KHE_INTERVAL_SET KheIntervalSetMake(HA_ARENA a)
{
  KHE_INTERVAL_SET res;
  HaMake(res, a);
  HaArrayInit(res->intervals, a);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheIntervalSetRead(char *str, KHE_INTERVAL_SET *res, HA_ARENA a)   */
/*                                                                           */
/*  Read an interval set from str, returning a pointer to the first place    */
/*  past what was read.                                                      */
/*                                                                           */
/*****************************************************************************/

static char *KheIntervalSetRead(char *str, KHE_INTERVAL_SET *res, HA_ARENA a)
{
  KHE_INTERVAL in;
  *res = KheIntervalSetMake(a);
  str = KheIntervalRead(str, &in);
  HaArrayAddLast((*res)->intervals, in);
  while( *str == ',' )
  {
    str = KheIntervalRead(str+1, &in);
    HaArrayAddLast((*res)->intervals, in);
  }
  return str;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheIntervalSetDebug(KHE_INTERVAL_SET is, FILE *fp)                  */
/*                                                                           */
/*  Debug printf of is onto fp.                                              */
/*                                                                           */
/*****************************************************************************/

static void KheIntervalSetDebug(KHE_INTERVAL_SET is, FILE *fp)
{
  KHE_INTERVAL in;  int i;
  HaArrayForEach(is->intervals, in, i)
  {
    if( i > 0 )
      fprintf(fp, ",");
    fprintf(fp, "%d-%d", in.first, in.last);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DAY_SELECT"                                               */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DAY_SELECT KheDaySelectMake(HA_ARENA a)                              */
/*                                                                           */
/*  Make a new, empty day select object.                                     */
/*                                                                           */
/*****************************************************************************/

static KHE_DAY_SELECT KheDaySelectMake(HA_ARENA a)
{
  KHE_DAY_SELECT res;
  HaMake(res, a);
  res->random_len = 0;
  HaArrayInit(res->interval_sets, a);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DAY_SELECT KheDaySelectMakeRandom(int random_len, HA_ARENA a)        */
/*                                                                           */
/*  Make a new, random day select object.                                    */
/*                                                                           */
/*****************************************************************************/

static KHE_DAY_SELECT KheDaySelectMakeRandom(int random_len, HA_ARENA a)
{
  KHE_DAY_SELECT res;  KHE_INTERVAL in;  KHE_INTERVAL_SET is;
  res = KheDaySelectMake(a);
  res->random_len = random_len;
  in = KheIntervalMake(0, 0);  /* placeholder, will be reset later */
  is = KheIntervalSetMake(a);
  HaArrayAddLast(is->intervals, in);
  HaArrayAddLast(res->interval_sets, is);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DAY_SELECT KheDaySelectMakeAll(KHE_FRAME days_frame,                 */
/*    int len, int inc, HA_ARENA a)                                          */
/*                                                                           */
/*  Make a day select object containing one interval set for each            */
/*  subsequence of len consecutive days, inc days apart; or fewer            */
/*  days if necessary at the end of the cycle.                               */
/*                                                                           */
/*****************************************************************************/

static KHE_DAY_SELECT KheDaySelectMakeAll(KHE_FRAME days_frame,
  int len, int inc, HA_ARENA a)
{
  KHE_INTERVAL in;  KHE_INTERVAL_SET is;  KHE_DAY_SELECT res;  int i, last;
  res = KheDaySelectMake(a);
  for( i = 0;  i < KheFrameTimeGroupCount(days_frame);  i += inc )
  {
    last = i + len - 1;
    if( last >= KheFrameTimeGroupCount(days_frame) )
      last = KheFrameTimeGroupCount(days_frame) - 1;
    in = KheIntervalMake(i, last);
    is = KheIntervalSetMake(a);
    HaArrayAddLast(is->intervals, in);
    HaArrayAddLast(res->interval_sets, is);
    if( last == KheFrameTimeGroupCount(days_frame) - 1 )
      break;
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheDaySelectRead(char *str, KHE_DAY_SELECT *res, HA_ARENA a)       */
/*                                                                           */
/*  Read a day select object from str, returning a pointer to the first      */
/*  place past what was read.                                                */
/*                                                                           */
/*****************************************************************************/

static char *KheDaySelectRead(char *str, KHE_DAY_SELECT *res, HA_ARENA a)
{
  KHE_INTERVAL_SET is;
  *res = KheDaySelectMake(a);
  str = KheIntervalSetRead(str, &is, a);
  HaArrayAddLast((*res)->interval_sets, is);
  while( *str == ';' )
  {
    str = KheIntervalSetRead(str+1, &is, a);
    HaArrayAddLast((*res)->interval_sets, is);
  }
  return str;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDaySelectDebug(KHE_DAY_SELECT ds, int verbosity,                 */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of ds onto fp with the given verbosity and indent.           */
/*                                                                           */
/*****************************************************************************/

static void KheDaySelectDebug(KHE_DAY_SELECT ds, int verbosity,
  int indent, FILE *fp)
{
  KHE_INTERVAL_SET is;  int i;
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  if( ds->random_len > 0 )
    fprintf(fp, "random:%d", ds->random_len);
  else
  {
    HaArrayForEach(ds->interval_sets, is, i)
    {
      if( i > 0 )
	fprintf(fp, ";");
      KheIntervalSetDebug(is, fp);
    }
  }
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KheDynamicResourceVLSNSolve"                                  */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddResources(KHE_DYNAMIC_RESOURCE_SOLVER drs,                 */
/*    ARRAY_KHE_RESOURCE *resources, int resource_count)                     */
/*                                                                           */
/*  Add the first resource_count resources from *resources to drs.           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddResources(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  ARRAY_KHE_RESOURCE *resources, int resource_count)
{
  int i;
  for( i = 0;  i < resource_count;  i++ )
    KheDynamicResourceSolverAddResource(drs, HaArray(*resources, i));
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddIntervalSet(KHE_DYNAMIC_RESOURCE_SOLVER drs,               */
/*    KHE_INTERVAL_SET is)                                                   */
/*                                                                           */
/*  Add is to drs.                                                           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddIntervalSet(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_INTERVAL_SET is)
{
  int i;  KHE_INTERVAL in;
  HaArrayForEach(is->intervals, in, i)
    KheDynamicResourceSolverAddDayRange(drs, in.first, in.last);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDaySelectChooseRandom(KHE_DAY_SELECT rs_drs_day_select,          */
/*    KHE_FRAME days_frame, KHE_RANDOM_GENERATOR *rgen)                      */
/*                                                                           */
/*  Choose a random interval of length rs_drs_day_select->random_len.        */
/*                                                                           */
/*****************************************************************************/

static void KheDaySelectChooseRandom(KHE_DAY_SELECT rs_drs_day_select,
  KHE_FRAME days_frame, KHE_RANDOM_GENERATOR *rgen)
{
  int tg_count, len, start_pos;  KHE_INTERVAL_SET is;  KHE_INTERVAL in;

  /* find start_pos */
  tg_count = KheFrameTimeGroupCount(days_frame);
  len = rs_drs_day_select->random_len;
  start_pos = KheRandomGeneratorNextRange(rgen, 0, tg_count - len);

  /* add [start_pos, start_pos + len - 1] to interval set */
  HnAssert(HaArrayCount(rs_drs_day_select->interval_sets) == 1,
    "KheDaySelectChooseRandom internal error 1");
  is = HaArrayFirst(rs_drs_day_select->interval_sets);
  HnAssert(HaArrayCount(is->intervals) == 1,
    "KheDaySelectChooseRandom internal error 2");
  in.first = start_pos;
  in.last = start_pos + len - 1;
  HaArrayPut(is->intervals, 0, in);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheSelectResources(ARRAY_KHE_RESOURCE *resources,                   */
/*    int resource_count, KHE_RANDOM_GENERATOR *rgen)                        */
/*                                                                           */
/*  Select resource_count resources randomly for adding to the solve,        */
/*  assuming that resource_count <= HaArrayCount(*resources).                */
/*                                                                           */
/*****************************************************************************/

static void KheSelectResources(ARRAY_KHE_RESOURCE *resources,
  int resource_count, KHE_RANDOM_GENERATOR *rgen)
{
  int i, pos;  KHE_RESOURCE r;
  for( i = 0;  i < resource_count;  i++ )
  {
    /* choose a random resource from position i onwards and swap it to i */
    pos = KheRandomGeneratorNextRange(rgen, i, HaArrayCount(*resources) - 1);
    HaArraySwap(*resources, i, pos, r);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheReadAllValue(char *str, KHE_FRAME days_frame, int *len, int *inc)*/
/*                                                                           */
/*  Read "all" value and set *len and *inc.                                  */
/*                                                                           */
/*****************************************************************************/

static bool KheReadAllValue(char *str, KHE_FRAME days_frame, int *len, int *inc)
{
  if( strstr(str, "all") != str )
    return *len = *inc = 0, false;
  switch( sscanf(str, "all:%d:%d", len, inc) )
  {
    case 0:

      /* failed to read both *len and *inc */
      *len = *inc = KheFrameTimeGroupCount(days_frame);
      break;

    case 1:

      /* read *len but failed to read *inc */
      *inc = *len;
      break;

    case 2:
    default:

      /* read both *len and *inc */
      break;
  }
  return true;
}
/* rs_drs_day_select = KheDaySelectMakeAll(days_frame, len, len, a); */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDynamicResourceVLSNSolve(KHE_SOLN soln, KHE_RESOURCE_TYPE rt,    */
/*    KHE_OPTIONS options)                                                   */
/*                                                                           */
/*  Carry out a VLSN search using calls to KheDynamicResourceSolverSolve.    */
/*                                                                           */
/*****************************************************************************/
static KHE_DRS_DOM_KIND KheDrsDomKindFromName(char *name);

bool KheDynamicResourceVLSNSolve(KHE_SOLN soln, KHE_RESOURCE_TYPE rt,
  KHE_OPTIONS options)
{
  KHE_DYNAMIC_RESOURCE_SOLVER drs;  KHE_COST init_cost;
  float rs_drs_time_limit;  KHE_TIMER timer;  KHE_INTERVAL_SET is;
  int rs_drs_solve_limit, rs_drs_daily_expand_limit, rs_drs_daily_prune_trigger;
  int rs_drs_resource_expand_limit, rs_drs_resource_count, i, len, inc, count;
  KHE_DRS_DOM_KIND rs_drs_dom_kind, rs_drs_cache_dom_kind;  char *str;
  bool priqueue, extra_selection, expand_by_shifts, cache, correlated_exprs;
  KHE_RANDOM_GENERATOR rgen;
  KHE_DAY_SELECT rs_drs_day_select;  HA_ARENA a;  KHE_FRAME days_frame;
  KHE_RESOURCE r;  ARRAY_KHE_RESOURCE resources;

  /* quit immediately if the rs_drs_off option is true */
  if( KheOptionsGetBool(options, "rs_drs_off", false) )
    return false;

  if( DEBUG1 )
    fprintf(stderr, "[ KheDynamicResourceVLSNSolve(%s, %s, options)\n",
      KheInstanceId(KheSolnInstance(soln)), KheResourceTypeId(rt));

  /* boilerplate */
  a = KheSolnArenaBegin(soln, false);
  days_frame = KheOptionsFrame(options, "gs_common_frame", soln);
  KheRandomGeneratorInit(&rgen, KheSolnDiversifier(soln));

  /* get rs_drs_time_limit and set timer */
  rs_drs_time_limit = KheTimeFromString(
    KheOptionsGet(options, "rs_drs_time_limit", "60"));
  if( rs_drs_time_limit >= 0.0 )
    timer = KheOptionsAddTimer(options, "vlsn", rs_drs_time_limit);
  else
    timer = NULL;

  /* get rs_drs_solve_limit */
  str = KheOptionsGet(options, "rs_drs_solve_limit", "1000");
  if( strcmp(str, "-") == 0 )
    rs_drs_solve_limit = INT_MAX;
  else
    rs_drs_solve_limit = KheOptionsGetInt(options, "rs_drs_solve_limit", 1000);

  /* get priqueue */
  priqueue = KheOptionsGetBool(options, "rs_drs_priqueue", false);

  /* get extra_selection */
  extra_selection =
    KheOptionsGetBool(options, "rs_drs_extra_selection", true);

  /* get expand_by_shifts */
  expand_by_shifts =
    KheOptionsGetBool(options, "rs_drs_expand_by_shifts", true);

  /* get rs_drs_correlated_exprs */
  correlated_exprs =
    KheOptionsGetBool(options, "rs_drs_correlated_exprs", false);

  /* get rs_drs_daily_expand_limit */
  rs_drs_daily_expand_limit =
    KheOptionsGetInt(options, "rs_drs_daily_expand_limit", 0);

  /* get rs_drs_daily_prune_trigger */
  rs_drs_daily_prune_trigger =
    KheOptionsGetInt(options, "rs_drs_daily_prune_trigger", 0);

  /* get rs_drs_resource_expand_limit */
  rs_drs_resource_expand_limit =
    KheOptionsGetInt(options, "rs_drs_resource_expand_limit", 0);

  /* get rs_drs_dom_kind */
  str = KheOptionsGet(options, "rs_drs_dom_kind", "indexedu");
  rs_drs_dom_kind = KheDrsDomKindFromName(str);

  /* get cache and rs_drs_cache_dom_kind */
  str = KheOptionsGet(options, "rs_drs_cache_dom_kind", "nocache");
  if( strcmp(str, "nocache") == 0 )
    cache = false, rs_drs_cache_dom_kind = KHE_DRS_DOM_LIST_NONE;
  else if( strcmp(str, "same") == 0 )
    cache = true, rs_drs_cache_dom_kind = rs_drs_dom_kind;
  else
    cache = true, rs_drs_cache_dom_kind = KheDrsDomKindFromName(str);

  /* get rs_drs_resource_count */
  rs_drs_resource_count = KheOptionsGetInt(options, "rs_drs_resource_count", 1);
  if( rs_drs_resource_count > KheResourceTypeResourceCount(rt) )
    rs_drs_resource_count = KheResourceTypeResourceCount(rt);

  /* get rs_drs_day_select */
  str = KheOptionsGet(options, "rs_drs_day_select", "all");
  if( is_digit(*str) )
  {
    /* interval sets given explicitly in string */
    str = KheDaySelectRead(str, &rs_drs_day_select, a);
    HnAssert(*str == '\0',
      "KheDynamicResourceVLSNSolve: error in rs_drs_day_select option");
  }
  else if( KheReadAllValue(str, days_frame, &len, &inc) )
  {
    /* interval sets spread along the cycle, with given length and increment */
    rs_drs_day_select = KheDaySelectMakeAll(days_frame, len, inc, a);
  }
  else if( strstr(str, "random:") == str )
  {
    if( sscanf(str, "random:%d", &len) != 1 )
      HnAbort("KheDynamicResourceVLSNSolve: error in rs_drs_day_select option");
    if( len > KheFrameTimeGroupCount(days_frame) )
      len = KheFrameTimeGroupCount(days_frame);
    rs_drs_day_select = KheDaySelectMakeRandom(len, a);
  }
  else
  {
    HnAbort("KheDynamicResourceVLSNSolve: error in rs_drs_day_select option");
    rs_drs_day_select = NULL;  /* keep compiler happy */
  }

  if( DEBUG2 )
  {
    fprintf(stderr, "  rs_drs_time_limit            %.2f\n", rs_drs_time_limit);
    fprintf(stderr, "  rs_drs_solve_limit           %d\n",  rs_drs_solve_limit);
    fprintf(stderr, "  rs_drs_correlated_exprs      %s\n",
      bool_show(correlated_exprs));
    fprintf(stderr, "  rs_drs_daily_expand_limit    %d\n",
      rs_drs_daily_expand_limit);
    fprintf(stderr, "  rs_drs_daily_prune_trigger   %d\n",
      rs_drs_daily_prune_trigger);
    fprintf(stderr, "  rs_drs_resource_expand_limit %d\n",
      rs_drs_resource_expand_limit);
    fprintf(stderr, "  rs_drs_dom_kind              %d\n",   rs_drs_dom_kind);
    fprintf(stderr, "  rs_drs_resource_count        %d\n",
      rs_drs_resource_count);
    fprintf(stderr, "  rs_drs_day_select          ");
    KheDaySelectDebug(rs_drs_day_select, 1, 0, stderr);
  }

  /* get resources */
  HaArrayInit(resources, a);
  for( i = 0;  i < KheResourceTypeResourceCount(rt);  i++ )
  {
    r = KheResourceTypeResource(rt, i);
    HaArrayAddLast(resources, r);
  }

  /* create solver */
  init_cost = KheSolnCost(soln);
  drs = KheDynamicResourceSolverMake(soln, rt, options);
  HnAssert(drs != NULL, "KheDynamicResourceVLSNSolve: unable to make solver");

  /* do the solves */
  count = 0;
  while( count < rs_drs_solve_limit && !KheOptionsTimeLimitReached(options) )
  {
    KheSelectResources(&resources, rs_drs_resource_count, &rgen);
    if( rs_drs_day_select->random_len > 0 )
      KheDaySelectChooseRandom(rs_drs_day_select, days_frame, &rgen);
    HaArrayForEach(rs_drs_day_select->interval_sets, is, i)
    {
      if( count >= rs_drs_solve_limit || KheOptionsTimeLimitReached(options) )
	break;
      KheDrsAddResources(drs, &resources, rs_drs_resource_count);
      KheDrsAddIntervalSet(drs, is);
      KheDynamicResourceSolverSolve(drs, priqueue, extra_selection,
	expand_by_shifts, correlated_exprs, rs_drs_daily_expand_limit,
	rs_drs_daily_prune_trigger, rs_drs_resource_expand_limit,
	rs_drs_dom_kind, cache, rs_drs_cache_dom_kind);
      count++;
    }
  }

  /* clean up and return */
  KheDynamicResourceSolverDelete(drs);
  KheSolnArenaEnd(soln, a);
  if( timer != NULL )
    KheOptionsDeleteTimer(options, timer);
  if( KheSolnCost(soln) < init_cost )
  {
    if( DEBUG1 )
      fprintf(stderr,
	"] KheDynamicResourceVLSNSolve returning true (%.5f < %.5f)\n",
	KheCostShow(KheSolnCost(soln)), KheCostShow(init_cost));
    return true;
  }
  else
  {
    if( DEBUG1 )
     fprintf(stderr, "] KheDynamicResourceVLSNSolve returning false\n");
    return false;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "Statistics"                                                   */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_STATS_DAY_INDEX KheDrsStatsDayIndexMake(int day_index,           */
/*    int made_count, int table_size, float running_time, HA_ARENA a)        */
/*                                                                           */
/*  Make a new KHE_DRS_STATS_DAY_INDEX object with these attributes.         */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_STATS_DAY_INDEX KheDrsStatsDayIndexMake(int day_index,
  int made_count, int table_size, float running_time, HA_ARENA a)
{
  KHE_DRS_STATS_DAY_INDEX res;
  HaMake(res, a);
  res->day_index = day_index;
  res->made_count = made_count;
  res->table_size = table_size;
  res->running_time = running_time;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_STATS_RESOURCE_COUNT KheDrsStatsResourceCountMake(               */
/*    int resource_count, HA_ARENA a)                                        */
/*                                                                           */
/*  Make a new KHE_DRS_STATS_RESOURCE_COUNT object with these attributes.    */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_STATS_RESOURCE_COUNT KheDrsStatsResourceCountMake(
  int resource_count, HA_ARENA a)
{
  KHE_DRS_STATS_RESOURCE_COUNT res;
  HaMake(res, a);
  res->resource_count = resource_count;
  HaArrayInit(res->days, a);
  res->cost = -1;  /* means no cost yet */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_STATS_RESOURCE_COUNT KheDrsStatsResourceCountBuild(              */
/*    int resource_count, KHE_DIM_DAY_INDEX day_index_dim, HA_ARENA a)       */
/*                                                                           */
/*  Build a complete resource count stats structure, with no data initially. */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_STATS_RESOURCE_COUNT KheDrsStatsResourceCountBuild(
  int resource_count, KHE_DIM_DAY_INDEX day_index_dim, HA_ARENA a)
{
  KHE_DRS_STATS_RESOURCE_COUNT res;  int day_index, i;
  KHE_DRS_STATS_DAY_INDEX dsd;
  res = KheDrsStatsResourceCountMake(resource_count, a);

  /* initial entry for before the first day */
  dsd = KheDrsStatsDayIndexMake(-1, -1, -1, -1.0, a);
  HaArrayAddLast(res->days, dsd);

  /* regular days */
  HaArrayForEach(day_index_dim->day_index_singles, day_index, i)
  {
    dsd = KheDrsStatsDayIndexMake(day_index, -1, -1, -1.0, a);
    HaArrayAddLast(res->days, dsd);
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_STATS_DOM_KIND KheDrsStatsDomKindMake(KHE_DRS_DOM_KIND dom_kind, */
/*    HA_ARENA a)                                                            */
/*                                                                           */
/*  Make a new KHE_DRS_STATS_DOM_KIND object with these attributes.          */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_STATS_DOM_KIND KheDrsStatsDomKindMake(
  KHE_DIM_DOM_KIND_ITEM dom_kind_item, HA_ARENA a)
{
  KHE_DRS_STATS_DOM_KIND res;
  HaMake(res, a);
  res->dom_kind_item = dom_kind_item;
  HaArrayInit(res->resource_counts, a);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_STATS_DOM_KIND KheDrsStatsDomKindBuild(                          */
/*    KHE_DRS_DOM_KIND dom_kind, KHE_DIM_RESOURCE_COUNT resource_count_dim,  */
/*    KHE_DIM_DAY_INDEX day_index_dim, HA_ARENA a)                           */
/*                                                                           */
/*  Build a complete dom kind stats structure, with no data initially.       */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_STATS_DOM_KIND KheDrsStatsDomKindBuild(
  KHE_DIM_DOM_KIND_ITEM dom_kind_item,
  KHE_DIM_RESOURCE_COUNT resource_count_dim,
  KHE_DIM_DAY_INDEX day_index_dim, HA_ARENA a)
{
  KHE_DRS_STATS_DOM_KIND res;  int resource_count, i;
  KHE_DRS_STATS_RESOURCE_COUNT dsr;
  res = KheDrsStatsDomKindMake(dom_kind_item, a);
  HaArrayForEach(resource_count_dim->resource_count_singles, resource_count, i)
  {
    dsr = KheDrsStatsResourceCountBuild(resource_count, day_index_dim, a);
    HaArrayAddLast(res->resource_counts, dsr);
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_STATS KheDrsStatsMake(HA_ARENA a)                                */
/*                                                                           */
/*  Make a new KHE_DRS_STATS object with these attributes.                   */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_STATS KheDrsStatsMake(HA_ARENA a)
{
  KHE_DRS_STATS res;
  HaMake(res, a);
  HaArrayInit(res->dom_kinds, a);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_STATS KheDrsStatsBuild(KHE_DIM_DOM_KIND dom_kind_dim,            */
/*    KHE_DIM_RESOURCE_COUNT resource_count_dim,                             */
/*    KHE_DIM_DAY_INDEX day_index_dim, HA_ARENA a)                           */
/*                                                                           */
/*  Build a complete stats structure, with no data initially.                */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_STATS KheDrsStatsBuild(KHE_DIM_DOM_KIND dom_kind_dim,
  KHE_DIM_RESOURCE_COUNT resource_count_dim,
  KHE_DIM_DAY_INDEX day_index_dim, HA_ARENA a)
{
  KHE_DRS_STATS res;  int i;  KHE_DIM_DOM_KIND_ITEM dom_kind_item;
  KHE_DRS_STATS_DOM_KIND dsk;
  res = KheDrsStatsMake(a);
  HaArrayForEach(dom_kind_dim->items, dom_kind_item, i)
  {
    dsk = KheDrsStatsDomKindBuild(dom_kind_item, resource_count_dim,
      day_index_dim, a);
    HaArrayAddLast(res->dom_kinds, dsk);
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheStatsAddCost(KHE_DRS_STATS stats, int i, int j, KHE_COST cost)   */
/*                                                                           */
/*  Add cost to stats, reporting an error if we find that two runs with      */
/*  the same resources and days but different dominance types produce        */
/*  solutions with different costs.                                          */
/*                                                                           */
/*  This function is only called when daily_expand_limit == 0 and            */
/*  daily_prune_trigger == 0, meaning unlimited in each case.  Otherwise,    */
/*  the check made here is not always applicable.                            */
/*                                                                           */
/*  Implementation note.  The common cost for the test with the j'th         */
/*  number of resources is held in the j'th element of the first element of  */
/*  stats.  This is the right thing to do; work out why yourself!            */
/*                                                                           */
/*****************************************************************************/

static void KheStatsAddCost(KHE_DRS_STATS stats, int j, KHE_COST cost)
{
  KHE_DRS_STATS_DOM_KIND dsk;  KHE_DRS_STATS_RESOURCE_COUNT dsr;
  dsk = HaArray(stats->dom_kinds, 0);
  dsr = HaArray(dsk->resource_counts, j);
  HnAssert(dsr->cost == -1 || dsr->cost == cost,
    "KheDynamicResourceVLSNTest: unequal costs for %d resources: %.5f vs %.5f",
    dsr->resource_count, KheCostShow(dsr->cost), KheCostShow(cost));
  dsr->cost = cost;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheStatsAdd(KHE_DRS_STATS stats, int i, int j, int k,               */
/*    int made_count, int table_size, float running_time)                    */
/*                                                                           */
/*  Add one pair of statistics to position (i, j, k) of stats.               */
/*                                                                           */
/*****************************************************************************/

static void KheStatsAdd(KHE_DRS_STATS stats, int i, int j, int k,
  int made_count, int table_size, float running_time)
{
  KHE_DRS_STATS_DOM_KIND dsk;  KHE_DRS_STATS_RESOURCE_COUNT dsr;
  KHE_DRS_STATS_DAY_INDEX dsd;

  dsk = HaArray(stats->dom_kinds, i);
  dsr = HaArray(dsk->resource_counts, j);
  dsd = HaArray(dsr->days, k);
  HnAssert(dsd->table_size == -1, "KheStatsAdd internal error");
  dsd->made_count = made_count;
  dsd->table_size = table_size;
  dsd->running_time = running_time;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheStatsGet(KHE_DRS_STATS stats, int i, int j, int k,               */
/*    int *made_count, int *table_size, float *running_time)                 */
/*                                                                           */
/*  Get one triple of statistics from position (i, j, k) of stats.           */
/*                                                                           */
/*****************************************************************************/

static void KheStatsGet(KHE_DRS_STATS stats, int i, int j, int k,
  int *made_count, int *table_size, float *running_time)
{
  KHE_DRS_STATS_DOM_KIND dsk;  KHE_DRS_STATS_RESOURCE_COUNT dsr;
  KHE_DRS_STATS_DAY_INDEX dsd;
  if( DEBUG5 )
    fprintf(stderr, "[ KheStatsGet(stats, %d, %d, %d, -, -)\n", i, j, k);

  HnAssert(0 <= i && i < HaArrayCount(stats->dom_kinds),
    "KheStatsGet internal error 1");
  dsk = HaArray(stats->dom_kinds, i);
  HnAssert(0 <= j && j < HaArrayCount(dsk->resource_counts),
    "KheStatsGet internal error 2");
  dsr = HaArray(dsk->resource_counts, j);
  HnAssert(0 <= k && k < HaArrayCount(dsr->days),
    "KheStatsGet internal error 3");
  dsd = HaArray(dsr->days, k);
  HnAssert(dsd->table_size != -1, "KheStatsGet internal error 4");
  *made_count = dsd->made_count;
  *table_size = dsd->table_size;
  *running_time = dsd->running_time;
  if( DEBUG5 )
    fprintf(stderr, "] KheStatsGet returning *made_count = %d, *table_size = "
      "%d, *running_time = %.2f\n", *made_count, *table_size, *running_time);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "reading the rs_drs_test option"                               */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  bool KheNextToken(char **str, char *delimiter, char **val)               */
/*                                                                           */
/*  Search *str for the next token, defined to be everything up to but       */
/*  not including delimiter, or if that is not found, then end of string.    */
/*  If there is a token present, set *val to it (this will be just the       */
/*  initial value of *str), update *str to point to the start of the next    */
/*  token, and return true.  Otherwise set *val to NULL and return false.    */
/*                                                                           */
/*  An empty token followed by a delimiter is returned in *val.  An          */
/*  empty token at the end of the string is ignored.                         */
/*                                                                           */
/*  And yes, I know that strtok_r does essentially the same thing.           */
/*                                                                           */
/*****************************************************************************/

static bool KheNextToken(char **str, char *delimiter, char **val)
{
  char *tmp;
  if( **str == '\0' )
  {
    /* we're already at end of string, so no token; return false */
    *val = NULL;
    return false;
  }
  else
  {
    *val = *str;
    tmp = strstr(*str, delimiter);
    if( tmp != NULL )
    {
      /* token is followed by delimiter */
      *tmp = '\0';
      *str = tmp + strlen(delimiter);
    }
    else
    {
      /* token is followed by end of string */
      *str += strlen(*str);
    }
    return true;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheParseOneOrRange(char **str, char **val1, char **val2)            */
/*                                                                           */
/*  Parse up to the next comma, finding either <val1>-<val2> or <val1>.  If  */
/*  the first, set *val1 and *val2 to the two parts and return true.  If     */
/*  the second, set both *val1 and *val2 to the sole part and return true.   */
/*  If neither, return false.  Move *str on to after the token.              */
/*                                                                           */
/*****************************************************************************/

static bool KheParseOneOrRange(char **str, char **val1, char **val2)
{
  char *val;

  /* read up to next comma or end of string, fail if can't */
  if( !KheNextToken(str, ",", &val) )
    return false;

  /* read the first part, fail if can't */
  if( !KheNextToken(&val, "-", val1) )
    return false;

  /* read the second part, succeed if can */
  if( KheNextToken(&val, "-", val2) )
    return true;

  /* have first part but not second, succeed with both parts the same */
  *val2 = *val1;
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheParseOneOrRangeInt(char **str, int *val1, int *val2)             */
/*                                                                           */
/*  Like KheParseOneOrRange, only returning integer values.                  */
/*                                                                           */
/*****************************************************************************/

static bool KheParseOneOrRangeInt(char **str, int *val1, int *val2)
{
  char *s1, *s2;
  return KheParseOneOrRange(str, &s1, &s2) &&
    sscanf(s1, "%d", val1) == 1 && sscanf(s2, "%d", val2) == 1;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_KIND KheDrsDomKindFromName(char *str)                        */
/*                                                                           */
/*  Convert str into a dom kind.                                             */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_DOM_KIND KheDrsDomKindFromName(char *str)
{
  if( strcmp(str, "list_none") == 0 )
    return KHE_DRS_DOM_LIST_NONE;
  else if( strcmp(str, "list_strong") == 0 )
    return KHE_DRS_DOM_LIST_STRONG;
  else if( strcmp(str, "list_tradeoff") == 0 )
    return KHE_DRS_DOM_LIST_TRADEOFF;
  else if( strcmp(str, "list_uniform") == 0 )
    return KHE_DRS_DOM_LIST_UNIFORM;
  else if( strcmp(str, "hash_weak") == 0 )
    return KHE_DRS_DOM_HASH_WEAK;
  else if( strcmp(str, "hash_medium") == 0 )
    return KHE_DRS_DOM_HASH_MEDIUM;
  else if( strcmp(str, "trie_strong") == 0 )
    return KHE_DRS_DOM_TRIE_STRONG;
  else if( strcmp(str, "trie_tradeoff") == 0 )
    return KHE_DRS_DOM_TRIE_TRADEOFF;
  else if( strcmp(str, "indexed_tradeoff") == 0 )
    return KHE_DRS_DOM_INDEXED_TRADEOFF;
  else if( strcmp(str, "indexed_uniform") == 0 )
    return KHE_DRS_DOM_INDEXED_UNIFORM;
  else
  {
    HnAbort("KheDrsDomKindFromName: unknown dom kind name (%s)", str);
    return KHE_DRS_DOM_LIST_NONE;  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheDomKindName(KHE_DRS_DOM_KIND dom_kind)                          */
/*                                                                           */
/*  Return the string name of dom_kind.                                      */
/*                                                                           */
/*****************************************************************************/

static char *KheDomKindName(KHE_DRS_DOM_KIND dom_kind)
{
  switch( dom_kind )
  {
    case KHE_DRS_DOM_LIST_NONE:		return "list_none";
    case KHE_DRS_DOM_LIST_STRONG:	return "list_strong";
    case KHE_DRS_DOM_LIST_TRADEOFF:	return "list_tradeoff";
    case KHE_DRS_DOM_LIST_UNIFORM:	return "list_uniform";
    case KHE_DRS_DOM_HASH_WEAK:		return "hash_weak";
    case KHE_DRS_DOM_HASH_MEDIUM:	return "hash_medium";
    case KHE_DRS_DOM_TRIE_STRONG:	return "trie_strong";
    case KHE_DRS_DOM_TRIE_TRADEOFF:	return "trie_tradeoff";
    case KHE_DRS_DOM_INDEXED_TRADEOFF:	return "indexed_tradeoff";
    case KHE_DRS_DOM_INDEXED_UNIFORM:	return "indexed_uniform";

    default:
      HnAbort("KheDomKindName internal error");
      return NULL;  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDimDomKindItemDebug(KHE_DIM_DOM_KIND_ITEM dom_kind_item,         */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug dom_kind_item.                                                     */
/*                                                                           */
/*****************************************************************************/

static void KheDimDomKindItemDebug(KHE_DIM_DOM_KIND_ITEM dom_kind_item,
  int verbosity, int indent, FILE *fp)
{
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  fprintf(fp, "DomKindItem(%s, %s, %s, %s",
    dom_kind_item->priqueue ? "priqueue" : "false",
    dom_kind_item->extra_selection ? "extra_selection" : "false",
    dom_kind_item->expand_by_shifts ? "expand_by_shifts" : "false",
    KheDomKindName(dom_kind_item->main_dom_kind));
  if( dom_kind_item->cache )
    fprintf(fp, ", %s", KheDomKindName(dom_kind_item->cache_dom_kind));
  fprintf(fp, ")");
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DIM_DOM_KIND_ITEM KheDimDomKindItemMake(bool priqueue,               */
/*    bool extra_selection, bool expand_by_shifts,                           */
/*    KHE_DRS_DOM_KIND main_dom_kind, bool cache,                            */
/*    KHE_DRS_DOM_KIND cache_dom_kind, HA_ARENA a)                           */
/*                                                                           */
/*  Make one dom kind item.                                                  */
/*                                                                           */
/*****************************************************************************/

static KHE_DIM_DOM_KIND_ITEM KheDimDomKindItemMake(bool priqueue,
  bool extra_selection, bool expand_by_shifts,
  KHE_DRS_DOM_KIND main_dom_kind, bool cache,
  KHE_DRS_DOM_KIND cache_dom_kind, HA_ARENA a)
{
  KHE_DIM_DOM_KIND_ITEM res;
  HaMake(res, a);
  res->priqueue = priqueue;
  res->extra_selection = extra_selection;
  res->expand_by_shifts = expand_by_shifts;
  res->main_dom_kind = main_dom_kind;
  res->cache = cache;
  res->cache_dom_kind = cache_dom_kind;
  if( DEBUG6 )
    KheDimDomKindItemDebug(res, 2, 2, stderr);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDimDomKindItemParse(char **dim_str,                              */
/*    KHE_DIM_DOM_KIND_ITEM *dom_kind_item, HA_ARENA a)                      */
/*                                                                           */
/*  Parse one dom kind dim item.                                             */
/*                                                                           */
/*****************************************************************************/

static bool KheDimDomKindItemParse(char **dim_str,
  KHE_DIM_DOM_KIND_ITEM *dom_kind_item, HA_ARENA a)
{
  char *val, *main_str, *cache_str;
  bool cache, priqueue, extra_selection, expand_by_shifts;
  KHE_DRS_DOM_KIND main_dom_kind, cache_dom_kind;
  if( !KheNextToken(dim_str, ",", &val) )
    return *dom_kind_item = NULL, false;

  /* get optional initial "P" meaning priqueue */
  priqueue = false;
  if( val[0] == 'P' )
    val++, priqueue = true;

  /* get optional initial "D" meaning extra_selection */
  extra_selection = false;
  if( val[0] == 'D' )
    val++, extra_selection = true;

  /* get optional initial "S" meaning expand_by_shifts */
  expand_by_shifts = false;
  if( val[0] == 'S' )
    val++, expand_by_shifts = true;

  if( strstr(val, "@") != NULL )
  {
    /* format is str@str so use cache */
    KheNextToken(&val, "@", &main_str);
    main_dom_kind = KheDrsDomKindFromName(main_str);
    if( !KheNextToken(&val, "@", &cache_str) )
      cache = true, cache_dom_kind = main_dom_kind;
    else if( strcmp(cache_str, "nocache") == 0 )
      cache = false, cache_dom_kind = KHE_DRS_DOM_LIST_NONE;
    else if( strcmp(cache_str, "same") == 0 )
      cache = true, cache_dom_kind = main_dom_kind;
    else
      cache = true, cache_dom_kind = KheDrsDomKindFromName(cache_str);
  }
  else
  {
    /* format is str so don't use cache */
    main_dom_kind = KheDrsDomKindFromName(val);
    cache = false, cache_dom_kind = KHE_DRS_DOM_LIST_NONE;
  }
  return *dom_kind_item = KheDimDomKindItemMake(priqueue,
    extra_selection, expand_by_shifts, main_dom_kind,
    cache, cache_dom_kind, a), true;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDimDomKindItemNameAdd(KHE_DIM_DOM_KIND_ITEM dom_kind_item,       */
/*    HA_ARRAY_NCHAR *name)                                                  */
/*                                                                           */
/*  Add the name of dom_kind_item to *name.                                  */
/*                                                                           */
/*****************************************************************************/

static void KheDimDomKindItemNameAdd(KHE_DIM_DOM_KIND_ITEM dom_kind_item,
  HA_ARRAY_NCHAR *name)
{
  if( dom_kind_item->priqueue )
    HnStringAdd(name, "%c", 'P');
  if( dom_kind_item->extra_selection )
    HnStringAdd(name, "%c", 'D');
  if( dom_kind_item->expand_by_shifts )
    HnStringAdd(name, "%c", 'S');
  HnStringAdd(name, "%s", KheDomKindName(dom_kind_item->main_dom_kind));
  if( dom_kind_item->cache )
  {
    HnStringAdd(name, "@");
    if( dom_kind_item->cache_dom_kind != dom_kind_item->main_dom_kind )
      HnStringAdd(name, KheDomKindName(dom_kind_item->cache_dom_kind));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DIM_DOM_KIND KheDimDomKindParse(char *dim_str, HA_ARENA a)           */
/*                                                                           */
/*  Parse a dominance kind dim object.                                       */
/*                                                                           */
/*****************************************************************************/

static KHE_DIM_DOM_KIND KheDimDomKindParse(char *dim_str, HA_ARENA a)
{
  KHE_DIM_DOM_KIND res;  KHE_DIM_DOM_KIND_ITEM dom_kind_item;

  if( DEBUG6 )
    fprintf(stderr, "[ KheDimDomKindParse(%s)\n", dim_str);
  /* skip the initial T and make res */
  dim_str++;
  HaMake(res, a);
  res->tag = KHE_DIM_DOM_KIND_TAG;
  HaArrayInit(res->items, a);

  /* parse the dominance type items */
  while( KheDimDomKindItemParse(&dim_str, &dom_kind_item, a) )
    HaArrayAddLast(res->items, dom_kind_item);

  /* all done */
  if( DEBUG6 )
    fprintf(stderr, "] KheDimDomKindParse returning\n");
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DIM_RESOURCE_COUNT KheDimResourceCountParse(char *dim_str,           */
/*    HA_ARENA a)                                                            */
/*                                                                           */
/*  Parse a resource count test dim object.                                  */
/*                                                                           */
/*****************************************************************************/

static KHE_DIM_RESOURCE_COUNT KheDimResourceCountParse(char *dim_str,
  HA_ARENA a)
{
  KHE_DIM_RESOURCE_COUNT res;  int val, val1, val2;

  /* skip the initial R and make res */
  dim_str++;
  HaMake(res, a);
  res->tag = KHE_DIM_RESOURCE_COUNT_TAG;
  HaArrayInit(res->resource_count_ranges, a);
  HaArrayInit(res->resource_count_singles, a);

  /* parse the resource counts */
  while( KheParseOneOrRangeInt(&dim_str, &val1, &val2) )
  {
    HaArrayAddLast(res->resource_count_ranges, val1);
    HaArrayAddLast(res->resource_count_ranges, val2);
    for( val = val1;  val <= val2;  val++ )
      HaArrayAddLast(res->resource_count_singles, val);
  }

  /* all done */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheOutputWholeAxisParse(char **str)                                 */
/*                                                                           */
/*  Parse the optional L denoting do not output whole axis.                  */
/*                                                                           */
/*****************************************************************************/

static bool KheOutputWholeAxisParse(char **str)
{
  if( **str == 'L' )
  {
    (*str)++;
    return false;
  }
  else
    return true;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DIM_DAY_INDEX KheDimDayIndexParse(char *dim_str,                     */
/*    KHE_FRAME days_frame, HA_ARENA a)                                      */
/*                                                                           */
/*  Parse a day index dim object.                                            */
/*                                                                           */
/*****************************************************************************/

static KHE_DIM_DAY_INDEX KheDimDayIndexParse(char *dim_str,
  KHE_FRAME days_frame, HA_ARENA a)
{
  KHE_DIM_DAY_INDEX res;  int val, val1, val2;
  KHE_TIME_GROUP tg;

  /* skip the initial D, parse the output whole axis flag, and make res */
  dim_str++;
  HaMake(res, a);
  res->tag = KHE_DIM_DAY_INDEX_TAG;
  res->output_whole_axis = KheOutputWholeAxisParse(&dim_str);
  HaArrayInit(res->day_time_group_ranges, a);
  HaArrayInit(res->day_index_ranges, a);
  HaArrayInit(res->day_index_singles, a);

  /* parse the resources flags */
  while( KheParseOneOrRangeInt(&dim_str, &val1, &val2) )
  {
    /* set day_time_group_ranges */
    tg = KheFrameTimeGroup(days_frame, val1);
    HaArrayAddLast(res->day_time_group_ranges, tg);
    tg = KheFrameTimeGroup(days_frame, val2);
    HaArrayAddLast(res->day_time_group_ranges, tg);

    /* set day_index_ranges */
    HaArrayAddLast(res->day_index_ranges, val1);
    HaArrayAddLast(res->day_index_ranges, val2);

    /* set day_index_singles */
    for( val = val1;  val <= val2;  val++ )
      HaArrayAddLast(res->day_index_singles, val);
  }

  /* all done */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_RUN KheDrsRunParse(char *run_str, KHE_SOLN soln,                 */
/*    KHE_RESOURCE_TYPE rt, KHE_FRAME days_frame, HA_ARENA a)                */
/*                                                                           */
/*  Create a KHE_DRS_RUN object by parsing run_str.                          */
/*                                                                           */
/*****************************************************************************/
static void KheDrsRunDebug(KHE_DRS_RUN run, int verbosity, int indent,
  FILE *fp);

static KHE_DRS_RUN KheDrsRunParse(char *run_str, KHE_SOLN soln,
  KHE_RESOURCE_TYPE rt, KHE_FRAME days_frame, HA_ARENA a)
{
  KHE_DRS_RUN res;  char *dim_str;  KHE_RESOURCE r;  int i;
  KHE_RANDOM_GENERATOR rgen;

  /* create the object and set the first few attributes */
  HaMake(res, a);
  res->arena = a;
  res->soln = soln;
  res->resource_type = rt;

  /* set resources - initialize and randomize the lot */
  HaArrayInit(res->resources, a);
  for( i = 0;  i < KheResourceTypeResourceCount(rt);  i++ )
  {
    r = KheResourceTypeResource(rt, i);
    HaArrayAddLast(res->resources, r);
  }
  KheRandomGeneratorInit(&rgen, KheSolnDiversifier(soln));
  KheSelectResources(&res->resources, HaArrayCount(res->resources), &rgen);

  /* set the dims; make sure they are all different and there are 3 of them */
  HaArrayInit(res->dims, a);
  res->dom_kind_dim = NULL;
  res->resource_count_dim = NULL;
  res->day_index_dim = NULL;
  while( KheNextToken(&run_str, ":", &dim_str) )
  {
    switch( *dim_str )
    {
      case 'T':
	
	HnAssert(res->dom_kind_dim == NULL,
	  "KheDrsRunParse: rs_drs_test run has two T dimensions");
	res->dom_kind_dim = KheDimDomKindParse(dim_str, a);
	HaArrayAddLast(res->dims, (KHE_DIM) res->dom_kind_dim);
	break;

      case 'R':
	
	HnAssert(res->resource_count_dim == NULL,
	  "KheDrsRunParse: rs_drs_test run has two R dimensions");
	res->resource_count_dim = KheDimResourceCountParse(dim_str, a);
	HaArrayAddLast(res->dims, (KHE_DIM) res->resource_count_dim);
	break;

      case 'D':
	
	HnAssert(res->day_index_dim == NULL,
	  "KheDrsRunParse: rs_drs_test run has two D dimensions");
	res->day_index_dim = KheDimDayIndexParse(dim_str, days_frame, a);
	HaArrayAddLast(res->dims, (KHE_DIM) res->day_index_dim);
	break;

      default:
	
	HnAbort("KheDrsRunParse: rs_drs_test run missing initial T, R, or D");
	break;

    }
  }
  HnAssert(HaArrayCount(res->dims) == 3, "KheRunTest: wrong number of "
    "dimensions (expected 3, found %d)", HaArrayCount(res->dims));

  /* set the stats and return */
  res->stats = KheDrsStatsBuild(res->dom_kind_dim,
    res->resource_count_dim, res->day_index_dim, a);
  if( DEBUG4 )
    KheDrsRunDebug(res, 2, 2, stderr);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDimDebug(KHE_DIM dim, int verbosity, int indent, FILE *fp)       */
/*                                                                           */
/*  Debug print of dim onto fp with the given verbosity and indent.          */
/*                                                                           */
/*****************************************************************************/

static void KheDimDebug(KHE_DIM dim, int verbosity, int indent, FILE *fp)
{
  KHE_DIM_DOM_KIND dimt;  KHE_DIM_RESOURCE_COUNT dimr;  KHE_DIM_DAY_INDEX dimd;
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  fprintf(fp, "[ ");
  switch( dim->tag )
  {
    case KHE_DIM_DOM_KIND_TAG:

      dimt = (KHE_DIM_DOM_KIND) dim;
      fprintf(fp, "KHE_DIM_DOM_KIND(%d items)", HaArrayCount(dimt->items));
      break;

    case KHE_DIM_RESOURCE_COUNT_TAG:

      dimr = (KHE_DIM_RESOURCE_COUNT) dim;
      fprintf(fp, "KHE_DIM_RESOURCE_COUNT(%d singles)",
	HaArrayCount(dimr->resource_count_singles));
      break;

    case KHE_DIM_DAY_INDEX_TAG:

      dimd = (KHE_DIM_DAY_INDEX) dim;
      fprintf(fp, "KHE_DIM_DAY_INDEX(%d singles)",
	HaArrayCount(dimd->day_index_singles));
      break;

    default:

      HnAbort("KheDimDebug internal error (tag %d)", dim->tag);
  }
  fprintf(fp, " ]");
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsRunDebug(KHE_DRS_RUN run, int verbosity, int indent, FILE *fp)*/
/*                                                                           */
/*  Debug print of run onto fp with the given verbosity and indent.          */
/*                                                                           */
/*****************************************************************************/

void KheDrsRunDebug(KHE_DRS_RUN run, int verbosity, int indent, FILE *fp)
{
  KHE_DIM dim;  int i;
  fprintf(fp, "%*s[ Run(%s, %s)\n", indent, "",
    KheInstanceId(KheSolnInstance(run->soln)),
    KheResourceTypeId(run->resource_type));
  HaArrayForEach(run->dims, dim, i)
    KheDimDebug(dim, verbosity, indent + 2, fp);
  fprintf(fp, "%*s]\n", indent, "");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "graph printing"                                               */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDimNameAdd(KHE_DIM dim, HA_ARRAY_NCHAR *name)                    */
/*                                                                           */
/*  Add the name of dim to *name.                                            */
/*                                                                           */
/*****************************************************************************/

static void KheDimNameAdd(KHE_DIM dim, HA_ARRAY_NCHAR *name)
{
  int i, val1, val2;
  KHE_DIM_DOM_KIND dimk;  KHE_DIM_DOM_KIND_ITEM dom_kind_item;
  KHE_DIM_RESOURCE_COUNT dimr;
  KHE_DIM_DAY_INDEX dimd;
  switch( dim->tag )
  {
    case KHE_DIM_DOM_KIND_TAG:

      dimk = (KHE_DIM_DOM_KIND) dim;
      HaArrayForEach(dimk->items, dom_kind_item, i)
      {
	if( i > 0 )
	  HnStringAdd(name, ",");
	KheDimDomKindItemNameAdd(dom_kind_item, name);
      }
      break;

    case KHE_DIM_RESOURCE_COUNT_TAG:

      dimr = (KHE_DIM_RESOURCE_COUNT) dim;
      HnStringAdd(name, "r");
      for( i = 0;  i < HaArrayCount(dimr->resource_count_ranges);  i += 2 )
      {
	if( i > 0 )
	  HnStringAdd(name, ",");
        val1 = HaArray(dimr->resource_count_ranges, i);
        val2 = HaArray(dimr->resource_count_ranges, i + 1);
	if( val1 == val2 )
	  HnStringAdd(name, "%d", val1);
	else
	  HnStringAdd(name, "%d-%d", val1, val2);
      }
      break;

    case KHE_DIM_DAY_INDEX_TAG:

      dimd = (KHE_DIM_DAY_INDEX) dim;
      HnStringAdd(name, "d");
      for( i = 0;  i < HaArrayCount(dimd->day_index_ranges);  i += 2 )
      {
	if( i > 0 )
	  HnStringAdd(name, ",");
        val1 = HaArray(dimd->day_index_ranges, i);
        val2 = HaArray(dimd->day_index_ranges, i + 1);
	if( val1 == val2 )
	  HnStringAdd(name, "%d", val1);
	else
	  HnStringAdd(name, "%d-%d", val1, val2);
      }
      break;

    default:

      HnAbort("KheDimName internal error");
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDimItemNameAdd(KHE_DIM dim, int i, HA_ARRAY_NCHAR *name)         */
/*                                                                           */
/*  Add the name of the i'th element of dim to *name.                        */
/*                                                                           */
/*****************************************************************************/

static void KheDimItemNameAdd(KHE_DIM dim, int i, HA_ARRAY_NCHAR *name)
{
  int val;  KHE_DIM_DOM_KIND_ITEM dom_kind_item;  KHE_DIM_DOM_KIND dimk;
  KHE_DIM_RESOURCE_COUNT dimr;  KHE_DIM_DAY_INDEX dimd;
  switch( dim->tag )
  {
    case KHE_DIM_DOM_KIND_TAG:

      dimk = (KHE_DIM_DOM_KIND) dim;
      dom_kind_item = HaArray(dimk->items, i);
      KheDimDomKindItemNameAdd(dom_kind_item, name);
      break;

    case KHE_DIM_RESOURCE_COUNT_TAG:

      dimr = (KHE_DIM_RESOURCE_COUNT) dim;
      val = HaArray(dimr->resource_count_singles, i);
      HnStringAdd(name, "r%d", val);
      break;

    case KHE_DIM_DAY_INDEX_TAG:

      dimd = (KHE_DIM_DAY_INDEX) dim;
      val = HaArray(dimd->day_index_singles, i);
      HnStringAdd(name, "d%d", val);
      break;

    default:

      HnAbort("KheDimItemNameAdd internal error");
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheDimItemLabel(KHE_DIM dim, int i, KHE_RESOURCE_TYPE rt,          */
/*    HA_ARENA a)                                                            */
/*                                                                           */
/*  Return the label of the i'th element of dim.                             */
/*                                                                           */
/*****************************************************************************/

static char *KheDimItemLabel(KHE_DIM dim, int i, KHE_RESOURCE_TYPE rt,
  HA_ARENA a)
{
  int val;  KHE_DIM_DOM_KIND_ITEM dom_kind_item;  KHE_DIM_DOM_KIND dimk;
  KHE_DIM_RESOURCE_COUNT dimr;  KHE_DIM_DAY_INDEX dimd;
  HA_ARRAY_NCHAR res;
  HnStringBegin(res, a);
  switch( dim->tag )
  {
    case KHE_DIM_DOM_KIND_TAG:

      dimk = (KHE_DIM_DOM_KIND) dim;
      dom_kind_item = HaArray(dimk->items, i);
      KheDimDomKindItemNameAdd(dom_kind_item, &res);
      break;

    case KHE_DIM_RESOURCE_COUNT_TAG:

      dimr = (KHE_DIM_RESOURCE_COUNT) dim;
      val = HaArray(dimr->resource_count_singles, i);
      HnStringAdd(&res, "%d %s resources", val, KheResourceTypeId(rt));
      break;

    case KHE_DIM_DAY_INDEX_TAG:

      dimd = (KHE_DIM_DAY_INDEX) dim;
      val = HaArray(dimd->day_index_singles, i);
      HnStringAdd(&res, "%d days", val);
      break;

    default:

      HnAbort("KheDimItemLabel internal error");
      break;
  }
  return HnStringEnd(res);
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheRunItemFileName(KHE_DRS_RUN drs_run, int i)                     */
/*                                                                           */
/*  Return a suitable file name for item i of drs_run.                       */
/*                                                                           */
/*****************************************************************************/

static char *KheRunItemFileName(KHE_DRS_RUN drs_run, int i, char *suffix)
{
  HA_ARRAY_NCHAR res;
  HnStringBegin(res, drs_run->arena);
  HnStringAdd(&res, "%s_%s_", KheInstanceId(KheSolnInstance(drs_run->soln)),
    KheResourceTypeId(drs_run->resource_type));
  KheDimItemNameAdd(HaArray(drs_run->dims, 0), i, &res);
  HnStringAdd(&res, "_");
  KheDimNameAdd(HaArray(drs_run->dims, 1), &res);
  HnStringAdd(&res, "_");
  KheDimNameAdd(HaArray(drs_run->dims, 2), &res);
  if( suffix != NULL )
    HnStringAdd(&res, "_%s", suffix);
  return HnStringEnd(res);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "main testing functions"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  char *KheBelowCaption(KHE_DIM dim, HA_ARENA a)                           */
/*                                                                           */
/*  Make a below caption for dim.                                            */
/*                                                                           */
/*****************************************************************************/

static char *KheDimBelowCaption(KHE_DIM dim, HA_ARENA a)
{
  KHE_DIM_DOM_KIND dimk;  KHE_DIM_RESOURCE_COUNT dimr;
  KHE_DIM_DAY_INDEX dimd;  HA_ARRAY_NCHAR anc;
  KHE_TIME_GROUP tg1, tg2;  int val1, val2, i;
  KHE_DIM_DOM_KIND_ITEM dom_kind_item;
  HnStringBegin(anc, a);
  switch( dim->tag )
  {
    case KHE_DIM_DOM_KIND_TAG:

      dimk = (KHE_DIM_DOM_KIND) dim;
      HnStringAdd(&anc, "Dominance kind (");
      HaArrayForEach(dimk->items, dom_kind_item, i)
      {
	if( i > 0 )
          HnStringAdd(&anc, ", ");
	KheDimDomKindItemNameAdd(dom_kind_item, &anc);
      }
      HnStringAdd(&anc, ")");
      break;

    case KHE_DIM_RESOURCE_COUNT_TAG:

      dimr = (KHE_DIM_RESOURCE_COUNT) dim;
      HnStringAdd(&anc, "Number of resources (");
      for( i = 0;  i < HaArrayCount(dimr->resource_count_ranges);  i += 2 )
      {
	if( i > 0 )
          HnStringAdd(&anc, ", ");
	val1 = HaArray(dimr->resource_count_ranges, i);
	val2 = HaArray(dimr->resource_count_ranges, i + 1);
	if( val1 == val2 )
          HnStringAdd(&anc, "%d", val1);
	else
          HnStringAdd(&anc, "%d-%d", val1, val2);
      }
      HnStringAdd(&anc, ")");
      break;

    case KHE_DIM_DAY_INDEX_TAG:

      dimd = (KHE_DIM_DAY_INDEX) dim;
      HnStringAdd(&anc, "Days (");
      for( i = 0;  i < HaArrayCount(dimd->day_time_group_ranges);  i += 2 )
      {
	if( i > 0 )
          HnStringAdd(&anc, ", ");
	tg1 = HaArray(dimd->day_time_group_ranges, i);
	tg2 = HaArray(dimd->day_time_group_ranges, i + 1);
	if( tg1 == tg2 )
          HnStringAdd(&anc, "%s", KheTimeGroupId(tg1));
	else
          HnStringAdd(&anc, "%s-%s", KheTimeGroupId(tg1), KheTimeGroupId(tg2));
      }
      HnStringAdd(&anc, ")");
      break;

    default:

      HnAbort("KheDimBelowCaption internal error");
      break;
  }
  return HnStringEnd(anc);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheIndexAssign(KHE_DIM dim, int index, int *dom_kind_index,         */
/*    int *resource_count_index, int *day_index_index)                       */
/*                                                                           */
/*  Assign index to *dom_kind_index, *resource_count_index, or               */
/*  *day_index_index depending on the type of dim.                           */
/*                                                                           */
/*****************************************************************************/

static void KheIndexAssign(KHE_DIM dim, int index, int *dom_kind_index,
  int *resource_count_index, int *day_index_index)
{
  switch( dim->tag )
  {
    case KHE_DIM_DOM_KIND_TAG:
      
      *dom_kind_index = index;
      break;

    case KHE_DIM_RESOURCE_COUNT_TAG:
				
      *resource_count_index = index;
      break;

    case KHE_DIM_DAY_INDEX_TAG:	
				
      *day_index_index = index;
      break;

    default:		

      HnAbort("KheIndexAssign internal error");
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheRunGetStats(KHE_DRS_RUN drs_run, int i, int j, int k,            */
/*    int *made_count, int *table_size, float *running_time)                 */
/*                                                                           */
/*  Get the stats at index (i, j, k).                                        */
/*                                                                           */
/*****************************************************************************/

static void KheRunGetStats(KHE_DRS_RUN drs_run, int i, int j, int k,
  int *made_count, int *table_size, float *running_time)
{
  int dom_kind_index, resource_count_index, day_index_index;

  /* sort out which of i, j, and k goes with which */
  dom_kind_index = resource_count_index = day_index_index = -1;
  KheIndexAssign(HaArray(drs_run->dims, 0), i,
    &dom_kind_index, &resource_count_index, &day_index_index);
  KheIndexAssign(HaArray(drs_run->dims, 1), j,
    &dom_kind_index, &resource_count_index, &day_index_index);
  KheIndexAssign(HaArray(drs_run->dims, 2), k,
    &dom_kind_index, &resource_count_index, &day_index_index);
  HnAssert(dom_kind_index != -1, "KheRunGetStats internal error 1");
  HnAssert(resource_count_index != -1, "KheRunGetStats internal error 2");
  HnAssert(day_index_index != -1, "KheRunGetStats internal error 3");

  /* get the stats */
  KheStatsGet(drs_run->stats, dom_kind_index, resource_count_index,
    day_index_index, made_count, table_size, running_time);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDimPrintRange(KHE_DIM dim, int *start, int *stop)                */
/*                                                                           */
/*  Return the range of indexes needing to be printed in dim.                */
/*                                                                           */
/*****************************************************************************/

static void KheDimPrintRange(KHE_DIM dim, int *start, int *stop)
{
  KHE_DIM_DOM_KIND dimk;  KHE_DIM_RESOURCE_COUNT dimr;  KHE_DIM_DAY_INDEX dimd;
  switch( dim->tag )
  {
    case KHE_DIM_DOM_KIND_TAG:

      dimk = (KHE_DIM_DOM_KIND) dim;
      *start = 0;
      *stop = HaArrayCount(dimk->items);
      break;

    case KHE_DIM_RESOURCE_COUNT_TAG:

      dimr = (KHE_DIM_RESOURCE_COUNT) dim;
      *start = 0;
      *stop = HaArrayCount(dimr->resource_count_singles);
      break;

    case KHE_DIM_DAY_INDEX_TAG:

      /******************************************************************/
      /*                                                                */
      /*  Implementation note.  The number of elements along the        */
      /*  day index axis is one greater than the length of the          */
      /*  dimd->day_index_singles array.  This is because the array     */
      /*  contains one element for each open day, but the axis has      */
      /*  one additional element, for the solution at the root of       */
      /*  the search tree, which does not correspond to any open day.   */
      /*                                                                */
      /******************************************************************/

      dimd = (KHE_DIM_DAY_INDEX) dim;
      if( dimd->output_whole_axis )
	*start = 0;
      else
	*start = HaArrayCount(dimd->day_index_singles);
      *stop = HaArrayCount(dimd->day_index_singles) + 1;
      break;

    default:

      HnAbort("KheDimPrintRange internal error");
      *start = *stop = 0;  /* keep compiler happy */
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheRunTest(KHE_SOLN soln, KHE_RESOURCE_TYPE rt, char *run_str,      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int daily_expand_limit,               */
/*    int daily_prune_trigger, KHE_FRAME days_frame, HA_ARENA a)             */
/*                                                                           */
/*  Perform one run.                                                         */
/*                                                                           */
/*****************************************************************************/

static void KheRunTest(KHE_SOLN soln, KHE_RESOURCE_TYPE rt, char *run_str,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int daily_expand_limit,
  int daily_prune_trigger, int resource_expand_limit,
  KHE_FRAME days_frame, HA_ARENA a)
{
  KHE_DRS_RUN drs_run;  KHE_DIM_DOM_KIND_ITEM dom_kind_item;
  int resource_count, i, j, k, first_day_index, last_day_index;
  int start1, stop1, start2, stop2, start3, stop3;
  int made_count, table_size;  float running_time;  KHE_DIM_DOM_KIND dim1;
  KHE_DIM_RESOURCE_COUNT dim2;  KHE_DIM_DAY_INDEX dim3;
  char *time_file_name, *made_file_name, *size_file_name;
  char *below_caption, *key_label, *label;
  KHE_DATASET kd_made, kd_size, kd_time;  KHE_DATASET_POINTS_TYPE points_type;
  KHE_FILE time_file, made_file, size_file;
  KHE_GRAPH made_graph, time_graph, size_graph;

  /* parse the dimensions, producing a run object */
  drs_run = KheDrsRunParse(run_str, soln, rt, days_frame, a);
  
  /* get the dimensions */
  dim1 = drs_run->dom_kind_dim;
  dim2 = drs_run->resource_count_dim;
  dim3 = drs_run->day_index_dim;

  /* run the dimensions */
  HaArrayForEach(dim1->items, dom_kind_item, i)
  {
    HaArrayForEach(dim2->resource_count_singles, resource_count, j)
    {
      if( resource_count <= HaArrayCount(drs_run->resources) )
      {
	/* add resource_count resources */
	KheDrsAddResources(drs, &drs_run->resources, resource_count);

	/* add day ranges (NB don't add singles, add ranges) */
	for( k = 0;  k < HaArrayCount(dim3->day_index_ranges);  k += 2 )
	{
	  first_day_index = HaArray(dim3->day_index_ranges, k);
	  last_day_index = HaArray(dim3->day_index_ranges, k + 1);
	  KheDynamicResourceSolverAddDayRange(drs, first_day_index,
	    last_day_index);
	}

	/* solve (without changing the solution) and extract statistics */
	KheDynamicResourceSolverTest(drs, dom_kind_item->priqueue,
          dom_kind_item->extra_selection, dom_kind_item->expand_by_shifts,
	  /* dom_kind_item->correlated_exprs */ false, daily_expand_limit,
	  daily_prune_trigger, resource_expand_limit,
	  dom_kind_item->main_dom_kind, dom_kind_item->cache,
	  dom_kind_item->cache_dom_kind);
	if( daily_expand_limit == 0 && daily_prune_trigger == 0 
	    && resource_expand_limit == 0 )
	  KheStatsAddCost(drs_run->stats, j, KheSolnCost(soln));
	  /* *** wrong! KheDynamicResourceSolverTest does not change soln *** */
	for( k = 0;  k < KheDynamicResourceSolverSolveStatsCount(drs);  k++ )
	{
	  KheDynamicResourceSolverSolveStats(drs, k, &made_count,
	    &table_size, &running_time);
	  KheStatsAdd(drs_run->stats, i, j, k, made_count,
	    table_size, running_time);
	}
      }
    }
  }

  /* print three graphs (made_count, table_size, and running time) */
  /* for each element of first dimension */
  KheDimPrintRange(HaArray(drs_run->dims, 0), &start1, &stop1);
  for( i = start1;  i < stop1;  i++ )
  {
    /* get the below caption and key label for these two graphs */
    below_caption = KheDimBelowCaption(HaArrayLast(drs_run->dims), a);
    key_label = KheDimItemLabel(HaArray(drs_run->dims, 0), i, rt,
      drs_run->arena);

    /* set up graph for table size */
    made_file_name = KheRunItemFileName(drs_run, i, "made");
    made_file = KheFileBegin(made_file_name, KHE_FILE_LOUT_STANDALONE);
    made_graph = KheGraphBegin(made_file);
    KheGraphSetBelowCaption(made_graph, below_caption);
    KheGraphSetLeftCaptionAndGap(made_graph, "Made", "1.0c");
    KheGraphSetKeyLabel(made_graph, key_label);

    /* set up graph for table size */
    size_file_name = KheRunItemFileName(drs_run, i, "size");
    size_file = KheFileBegin(size_file_name, KHE_FILE_LOUT_STANDALONE);
    size_graph = KheGraphBegin(size_file);
    KheGraphSetBelowCaption(size_graph, below_caption);
    KheGraphSetLeftCaptionAndGap(size_graph, "Size", "1.0c");
    KheGraphSetKeyLabel(size_graph, key_label);

    /* set up graph for running time */
    time_file_name = KheRunItemFileName(drs_run, i, "time");
    time_file = KheFileBegin(time_file_name, KHE_FILE_LOUT_STANDALONE);
    time_graph = KheGraphBegin(time_file);
    KheGraphSetBelowCaption(time_graph, below_caption);
    KheGraphSetLeftCaptionAndGap(time_graph, "Secs", "1.0c");
    KheGraphSetKeyLabel(time_graph, key_label);

    /* print one data set for each element of the second dimension */
    KheDimPrintRange(HaArray(drs_run->dims, 1), &start2, &stop2);
    for( j = start2;  j < stop2;  j++ )
    {
      /* label and points type */
      label = KheDimItemLabel(HaArray(drs_run->dims, 1), j, rt, a);
      points_type = (KHE_DATASET_POINTS_TYPE) (j % 10 + 1);

      /* made, size, and time graph data sets */
      kd_made = KheDataSetAdd(made_graph, points_type,
	KHE_DATASET_PAIRS_DASHED, label);
      kd_size = KheDataSetAdd(size_graph, points_type,
	KHE_DATASET_PAIRS_DASHED, label);
      kd_time = KheDataSetAdd(time_graph, points_type,
	KHE_DATASET_PAIRS_DASHED, label);
      KheDimPrintRange(HaArray(drs_run->dims, 2), &start3, &stop3);
      for( k = start3;  k < stop3;  k++ )
      {
	KheRunGetStats(drs_run, i, j, k, &made_count, &table_size,
	  &running_time);
	if( stop3 <= 28 || k % 2 == 0 || k == stop3 - 1 )
	{
	  KhePointAdd(kd_made, (float) k, (float) made_count);
	  KhePointAdd(kd_size, (float) k, (float) table_size);
	  KhePointAdd(kd_time, (float) k, running_time);
	}
      }
    }

    /* end graphs and files */
    KheGraphEnd(made_graph);
    KheFileEnd(made_file);
    KheGraphEnd(size_graph);
    KheFileEnd(size_file);
    KheGraphEnd(time_graph);
    KheFileEnd(time_file);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceVLSNTest(KHE_SOLN soln, KHE_RESOURCE_TYPE rt,     */
/*    KHE_OPTIONS options)                                                   */
/*                                                                           */
/*  Test dynamic resource solving and produce graphs of the results, as      */
/*  documented in the User's Guide.                                          */
/*                                                                           */
/*****************************************************************************/

void KheDynamicResourceVLSNTest(KHE_SOLN soln, KHE_RESOURCE_TYPE rt,
  KHE_OPTIONS options)
{
  HA_ARENA a;  char *rs_drs_test, *run;  KHE_DYNAMIC_RESOURCE_SOLVER drs;
  int daily_expand_limit, daily_prune_trigger, resource_expand_limit;
  KHE_FRAME days_frame;
  rs_drs_test = KheOptionsGet(options, "rs_drs_test", "none");
  if( strcmp(rs_drs_test, "none") != 0 )
  {
    /* find and execute each run */
    if( DEBUG3 )
    {
      fprintf(stderr, "[ KheDynamicResourceVLSNTest(%s, %s, options)\n",
	KheInstanceId(KheSolnInstance(soln)), KheResourceTypeId(rt));
      fprintf(stderr, "  %s\n", rs_drs_test);
    }
    daily_expand_limit =
      KheOptionsGetInt(options, "rs_drs_daily_expand_limit", 0);
    daily_prune_trigger =
      KheOptionsGetInt(options, "rs_drs_daily_prune_trigger", 0);
    resource_expand_limit =
      KheOptionsGetInt(options, "rs_drs_resource_expand_limit", 0);
    days_frame = KheOptionsFrame(options, "gs_common_frame", soln);
    a = KheSolnArenaBegin(soln, false);
    rs_drs_test = HnStringCopy(rs_drs_test, a);
    drs = KheDynamicResourceSolverMake(soln, rt, options);
    while( KheNextToken(&rs_drs_test, ";", &run) )
      KheRunTest(soln, rt, run, drs, daily_expand_limit,
	daily_prune_trigger, resource_expand_limit, days_frame, a);
    KheDynamicResourceSolverDelete(drs);
    KheSolnArenaEnd(soln, a);
    if( DEBUG3 )
      fprintf(stderr, "] KheDynamicResourceVLSNTest\n");
  }
}
